#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  mayhem/windll/advapi32.py
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are
#  met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following disclaimer
#    in the documentation and/or other materials provided with the
#    distribution.
#  * Neither the name of the project nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#  'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
#  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
#  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,rr
#  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
#  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import ctypes

from . import kernel32 as m_k32
import subprocplus.datatypes.windows as wintypes

TOKEN_ASSIGN_PRIMARY    = 0x0001
TOKEN_DUPLICATE         = 0x0002
TOKEN_IMPERSONATE       = 0x0004
TOKEN_QUERY             = 0x0008
TOKEN_QUERY_SOURCE      = 0x0010
TOKEN_ADJUST_PRIVILEGES = 0x0020
TOKEN_ADJUST_GROUPS     = 0x0040
TOKEN_ADJUST_DEFAULT    = 0x0080
TOKEN_ADJUST_SESSIONID  = 0x0100
TOKEN_ALL_ACCESS        = 0xf00ff

_advapi32 = ctypes.windll.advapi32

# https://learn.microsoft.com/en-us/windows/win32/api/securitybaseapi/nf-securitybaseapi-duplicatetokenex
DuplicateTokenEx = m_k32._patch_winfunctype(
	_advapi32.DuplicateTokenEx,
	wintypes.BOOL,
	(wintypes.HANDLE, wintypes.DWORD, wintypes.PSECURITY_ATTRIBUTES, wintypes.SECURITY_IMPERSONATION_LEVEL, wintypes.TOKEN_TYPE, wintypes.PHANDLE)
)

# https://learn.microsoft.com/en-us/windows/win32/api/securitybaseapi/nf-securitybaseapi-impersonateloggedonuser
ImpersonateLoggedOnUser = m_k32._patch_winfunctype(
	_advapi32.ImpersonateLoggedOnUser,
	wintypes.BOOL,
	(wintypes.HANDLE,)
)

# https://docs.microsoft.com/en-us/windows/win32/api/processthreadsapi/nf-processthreadsapi-openprocesstoken
OpenProcessToken = m_k32._patch_winfunctype(
	_advapi32.OpenProcessToken,
	wintypes.BOOL,
	(wintypes.HANDLE, wintypes.DWORD, wintypes.PHANDLE)
)

# https://docs.microsoft.com/en-us/windows/win32/api/processthreadsapi/nf-processthreadsapi-openthreadtoken
OpenThreadToken = m_k32._patch_winfunctype(
	_advapi32.OpenThreadToken,
	wintypes.BOOL,
	(wintypes.HANDLE, wintypes.DWORD, wintypes.BOOL, wintypes.PHANDLE)
)

# https://learn.microsoft.com/en-us/windows/win32/api/processthreadsapi/nf-processthreadsapi-setthreadtoken
SetThreadToken = m_k32._patch_winfunctype(
	_advapi32.SetThreadToken,
	wintypes.BOOL,
	(wintypes.PHANDLE, wintypes.HANDLE)
)

address = m_k32.GetModuleHandleW('advapi32.dll')
